##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Implement the CollectorNG issue-container content type."""

import os, re, traceback
from Globals import InitializeClass, DTMLFile, package_home

from AccessControl import ClassSecurityInfo, ModuleSecurityInfo, Permission
from AccessControl import getSecurityManager

from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl
from Products.CMFCore.PortalContent import PortalContent
from Products.CMFCore.WorkflowCore import WorkflowAction
from Products.CMFCore.CatalogTool import CatalogTool

from Products.CMFDefault.SkinnedFolder import SkinnedFolder
from Products.PythonScripts.PythonScript import PythonScript

from config.issue_config import all_issue_properties
from config.collector_config import all_collector_properties
from CollectorNGProperties  import CollectorNGPropertyManager
from types import UnicodeType
from Transcript import Transcript, TranscriptEntry

import util
from types import ListType

# Import permission names
from Products.CMFCore import CMFCorePermissions
from CollectorNGPermissions import *
from CollectorNGIssue import addCollectorNGIssue, CollectorNGIssue
from XML4CollectorNG import CollectorXML

INTERNAL_CATALOG_ID = 'collector_catalog'

# Factory type information -- makes Events objects play nicely
# with the Types Tool (portal_types)

factory_type_information = (
    {'id': 'CollectorNG',
     'content_icon': 'collector_icon.gif',
     'meta_type': 'CMF CollectorNG',
     'description': ('A CollectorNG is a facility for tracking bug reports and'
                     ' other issues.'), 
     'product': 'CMFCollectorNG',
     'factory': 'addCollectorNG',
     'allowed_content_types': ('CollectorNGIssue',
                               'CollectorNGCatalog',
                               'CollectorNG Subset',
                              ), 
     'immediate_view': 'collector_edit_form',
     'actions': ({'id': 'view',
                  'name': 'Browse',
                  'action': 'collector_contents',
                  'permissions': (ViewCollector,)},
                 {'id': 'addissue',
                  'name': 'New Issue',
                  'action': 'collector_add_issue_form',
                  'permissions': (AddCollectorIssue,)},
                 {'id': 'edit',
                  'name': 'Configure',
                  'action': 'collector_edit_form',
                  'permissions': (ManageCollector,)},
                 {'id': 'edit',
                  'name': 'Staff',
                  'action': 'collector_staff_form',
                  'permissions': (ManageCollector,)},
                 {'id': 'editproperties',
                  'name': 'History',
                  'action': 'collector_transcript_form',
                  'permissions': (ManageCollector,)},
                 {'id': 'editnotifications',
                  'name': 'Notifications',
                  'action': 'collector_notifications_form',
                  'permissions': (ManageCollector,)},
                 {'id': 'editissueproperties',
                  'name': 'Issue Properties',
                  'action': 'collector_issue_properties_form',
                  'permissions': (ManageCollector,)},
                 ),
     },
    )

_dtmldir = os.path.join(package_home(globals()), 'dtml')
addCollectorNGForm = DTMLFile('addCollectorNGForm', _dtmldir, Kind='CMF CollectorNG')

class CollectorNG(SkinnedFolder, CollectorXML):
    """Collection of IssueBundles."""

    meta_type = 'CMF CollectorNG'
    effective_date = expiration_date = None
    __allow_access_to_unprotected_subobjects__ = 1
    security = ClassSecurityInfo()

    __ac_roles__ = ('Manager', 'TrackerAdmin', 'Supporter','Reporter')
    # state_email - a dictionary pairing state names with email destinations
    # for all notifications occuring within that state.
    state_email = {}

    _properties=({'id':'title', 'type': 'string', 'mode':'w'},
                 {'id':'last_issue_id', 'type': 'int', 'mode':'w'},
                 {'id':'batch_size', 'type': 'int', 'mode':'w'},
                 )

    def __init__(self, id, title=''):

        SkinnedFolder.__init__(self, id, title)

        self._setup_internal_catalog()
        self.last_issue_id = 0
        self.transcript = Transcript()
        self.issuePropertyManager = CollectorNGPropertyManager(all_issue_properties)
        self.collectorPropertyManager = CollectorNGPropertyManager(all_collector_properties)

        for prop in self.collectorPropertyManager.getProperties():
            setattr(self, prop.getId(), prop.getDefaultValue())

        username = getSecurityManager().getUser().getUserName()
        util.add_local_role(self, username, 'Manager')
        util.add_local_role(self, username, 'TrackerAdmin')
        util.add_local_role(self, username, 'Owner')

        self._adjust_staff_roles(no_reindex=1)

        t_entry = TranscriptEntry(username)
        t_entry.addComment('Tracker created')
        self.transcript.add(t_entry)
        self.transcript._p_changed = 1
        self.issuePropertyManager._p_changed = 1

        return self

    def _setup_internal_catalog(self):
        """Create and situate properly configured collector catalog."""
        catalog = CollectorNGCatalog()
        self._setObject(catalog.id, catalog)
        catalog = catalog.__of__(self)

    security.declareProtected(CMFCorePermissions.View, 'get_internal_catalog')
    def get_internal_catalog(self):
        """ """
        return self._getOb(INTERNAL_CATALOG_ID)
        

    security.declareProtected(AddCollectorIssue, 'new_issue_id')
    def new_issue_id(self):
        """Return a new issue id, incrementing the internal counter."""
        lastid = self.last_issue_id = self.last_issue_id + 1
        return str(lastid)

    security.declareProtected(AddCollectorIssue, 'add_issue')
    def add_issue(self, properties, upload=None, reference=None):
           
        """Create a new collector issue."""

        id = self.new_issue_id()
        submitter_id = getSecurityManager().getUser().getUserName()

        setattr(properties, 'submitter_id', submitter_id)
        
        err = addCollectorNGIssue(self, id, properties, upload=upload, reference=reference)
        return id, err


    security.declareProtected(ManageCollector, 'edit')
    def edit(self, properties):

        userid = getSecurityManager().getUser().getUserName()
        t_entry = TranscriptEntry(userid)

        for key,new in properties.items():

            if isinstance(new, ListType):
                new = filter(None, new)   # filter out empty entries
            
            old = getattr(self, key) 

            if new != old:
                t_entry.addChange(key, old, new)
                setattr(self, key, new)

        if not self.email:
            raise ValueError, 'The tracker must have a valid email address'

        if not util.isValidEmailAddress(self.email):
            raise ValueError, 'Invalid email address: %s' % self.email

        if not self.managers:
            raise ValueError, "The manager list must not be empty"

        try:        
            self._adjust_participation_mode(properties.participation)
        except:
            self._adjust_participation_mode(self.participation)

        self._adjust_staff_roles()
        self.transcript.add(t_entry)
        self.transcript._p_changed = 1

        return str(t_entry)

    def _adjust_staff_roles(self, no_reindex=0):
        """Adjust local-role assignments to track staff roster settings.
        Ie, ensure: only designated supporters and managers have 'Reviewer'
        local role, only designated managers have 'Manager' local role.

        We reindex the issues if any local role changes occur, so
        allowedRolesAndUsers catalog index tracks.

        We return a list of changes (or non-changes)."""

        managers = self.managers
        supporters = self.supporters
        reporters = self.reporters
        changed = 0

        if not managers:
            # Something is awry.  Managers are not allowed to remove
            # themselves from the managers roster, and only managers should be
            # able to adjust the roles, so:
            changed = 1
            self.managers = managers = [getSecurityManager().getUser().getUserName()]

        if util.users_for_local_role(self, managers, 'TrackerAdmin'):
            changed = 1
        if util.users_for_local_role(self, supporters, 'Supporter'):
            changed = 1
        if util.users_for_local_role(self, reporters, 'Reporter'):
            changed = 1

        if changed and not no_reindex:
            self._reindex_issues()


    def _adjust_participation_mode(self, mode):
        """Set role privileges according to participation mode."""

        target_roles = ['Supporter','TrackerAdmin','Reporter', 'Manager', 'Owner']

        if mode == 'authenticated':
            target_roles = target_roles + ['Authenticated']
        elif mode == 'anyone':
            target_roles = target_roles + ['Authenticated', 'Anonymous']

        self.manage_permission(AddCollectorIssue,
                               roles=target_roles,
                               acquire=0)

        self.manage_permission(AddCollectorIssueFollowup,
                               roles=target_roles,
                               acquire=0)

        self.participation = mode

    security.declareProtected(ManageCollector, 'reinstate_catalog')
    def reinstate_catalog(self, internal_only=1):
        """Recreate and reload internal catalog, to accommodate drastic
        changes."""
        try:
            self._delObject(INTERNAL_CATALOG_ID)
        except AttributeError:
            pass
        self._setup_internal_catalog()
        self._reindex_issues(internal_only=internal_only)

    security.declareProtected(ManageCollector, 'update_states_email')
    def update_states_email(self, states_email):        
        """ update the mapping state -> email notification """

        for state in states_email.keys():
            email = states_email[state]

            if email:
                if not util.isValidEmailAddress(email):
                    raise ValueError, 'Invalid email address: %s' % email

                self.state_email[state] = email
            else:
                if self.state_email.has_key(email):
                    del self.state_email[k]


    def _reindex_issues(self, internal_only=1):
        """For, eg, allowedRolesAndUsers recompute after local_role changes.

        We also make sure that the AddCollectorNGIssueFollowup permission
        acquires (old workflows controlled this).  This isn't exactly the
        right place, but it is an expedient one."""

        for i in self.objectValues(spec='CMF CollectorNG Issue'):

            # Ensure the issue acquires AddCollectorNGIssueFollowup permission.
            for m in i.ac_inherited_permissions(1):
                if m[0] == AddCollectorIssueFollowup:
                    perm = Permission.Permission(m[0], m[1], i)
                    roles = perm.getRoles()
                    if type(roles) == type(()):
                        perm.setRoles(list(roles))

            i.reindexObject(internal_only=internal_only)


    security.declareProtected(CMFCorePermissions.View, 'Subject')
    def Subject(self):
        return self.issuePropertyManager.getPropertyById('topic').getValues().split('|')

    security.declareProtected(CMFCorePermissions.View, 'length')
    def length(self):
        """Use length protocol."""
        return self.__len__()


    security.declareProtected(AddCollectorIssue, "showNewIssueToken")
    def showNewIssueToken(self):
        """ display the email new issue token """

        if not self.email_submissions or len(self.newissue_token)==0:
            raise RuntimeError, "Email submissions are disabled"

        return "The token for submitting new issues through email is: %s" % self.newissue_token


    security.declareProtected(ManageCollector, 'issue_states')
    def issue_states(self):
        """ return a sorted sequence of states """

        # This looks like a hack but since it depends on the workflow name but
        # maybe cleaner than generating a stub as in CMFCollector :-)
        return self.portal_workflow.collectorng_issue_workflow.states._mapping.keys()

        
    def __len__(self):
        """length() protocol method."""
        return len(self.objectIds('CMF CollectorNG Issue')) 

    def __repr__(self):
        return ("<%s %s (%d issues) at 0x%s>"
                % (self.__class__.__name__, `self.id`, len(self),
                   hex(id(self))[2:]))

    def getCollectorTitle(self):
        """ return title of collector """
        return self.title

    def getCollectorDescription(self):
        """ return description of collector """
        return self.description


    def getSubtopics(self):
        """ return a sequence of tuples (topic, [subtopics for topic]) """

        reg = re.compile('(\((.*?):(.*?)\))')
        fields = reg.findall(self.issuePropertyManager.getPropertyById('subtopic').getValues())
        return [ (f[1], f[2].split('|'))  for f in fields] 


    def trans(self, text):
        """ translate a string using the translation_service """

        if not hasattr(self, "have_ts"):
            ts = getattr(self, "translation_service", None)
            if ts: self.have_ts = 1
            else:  self.have_ts = 0

        if not self.have_ts: return text

        if not hasattr(self, "_v_tsdomain"):
            self._v_tsdomain = getattr(self, "translation_service").getDomain("")

        res = self._v_tsdomain.translate(text)
        res = res or text
        if isinstance(res, UnicodeType):
            res = res.encode('iso-8859-15')

        return res
    

    security.declareProtected(ManageCollector, 'migrate')
    def migrate(self):
        """ migrate a CMFCollectorNG to new properties """

        updated = []
        for prop in all_collector_properties:
            if not self.collectorPropertyManager.hasProperty(prop.getId()):
                self.collectorPropertyManager.addProperty(prop)
                setattr(self, prop.getId(), prop.getDefaultValue())
                updated.append(prop.getId())

        install_proxied_scripts(self)
        self._p_changed = 1
        self.collectorPropertyManager._p_changed = 1

        if len(updated) > 0:                
            return 'CMFCollectorNG instance updated'        
        else:
            return 'CMFCollector instance is up-to-date'

InitializeClass(CollectorNG)
    
catalog_factory_type_information = (
    {'id': 'CollectorNG Catalog',
     'content_icon': 'collector_icon.gif',
     'meta_type': 'CMF CollectorNG Catalog',
     'description': ('Internal catalog.'), 
     'product': 'CMFCollectorNG',
     'factory': None,
     'immediate_view': None},)

class CollectorNGCatalog(CatalogTool):

    id = INTERNAL_CATALOG_ID
    meta_type = 'CMF CollectorNG Catalog'
    portal_type = 'CollectorNG Catalog'

    def enumerateIndexes(self):
        standard = CatalogTool.enumerateIndexes(self)
        custom = (('status', 'FieldIndex'),
                  ('topic', 'FieldIndex'),
                  ('subtopic', 'FieldIndex'),
                  ('classification', 'FieldIndex'),
                  ('importance', 'FieldIndex'),
                  ('security_related', 'FieldIndex'),
                  ('confidential', 'FieldIndex'),
                  ('submitter_id', 'FieldIndex'),
                  ('submitter_email', 'FieldIndex'),
                  ('version_info', 'TextIndex'),
                  ('assigned_to', 'KeywordIndex'),
                  ('deadline', 'FieldIndex'),
                  ('progress', 'FieldIndex'),
                  ('getId', 'FieldIndex'),
                  )
        custom = tuple([col for col in custom if col not in standard])
        return standard + custom

    def enumerateColumns( self ):
        """Return field names of data to be cached on query results."""
        standard = CatalogTool.enumerateColumns(self)
        custom = ('id',
                  'status',
                  'submitter_id',
                  'topic',
                  'subtopic',
                  'classification',
                  'importance',
                  'security_related',
                  'confidential',
                  'version_info',
                  'assigned_to',
                  'deadline',
                  'progress',
                  'getId',
                  )
        custom = tuple([col for col in custom if col not in standard])
        return standard + custom

InitializeClass(CollectorNGCatalog)


def addCollectorNG(self, id, title='', REQUEST=None):
    """
    Create a collector.
    """
    it = CollectorNG(id, title=title)
    self._setObject(id, it)
    it = it.__of__(self)

    it._addRole('Supporter')
    it._addRole('TrackerAdmin')
    it._addRole('Reporter')

    it = self._getOb(id)
    it._setPortalTypeName('CollectorNG')

    it.manage_permission(ManageCollector, roles=['Manager', 'TrackerAdmin', 'Owner'],
                         acquire=0)
    it.manage_permission(EditCollectorIssue,
                         roles=['Manager', 'TrackerAdmin', 'Owner'],
                         acquire=0)
    it.manage_permission(EditCollectorIssueAsSupporter,
                         roles=['Manager', 'TrackerAdmin', 'Supporter', 'Owner'],
                         acquire=0)
    it.manage_permission(AddCollectorIssue,
                         roles=['TrackerAdmin','Supporter','Reporter', 'Manager', 'Owner'],
                         acquire=0)
    it.manage_permission(AddCollectorIssueFollowup,
                         roles=['TrackerAdmin','Supporter', 'Manager' ,'Reporter'],
                         acquire=0)
    it.manage_permission(CMFCorePermissions.AccessInactivePortalContent,
                         roles=['TrackerAdmin','Supporter','Reporter', 'Manager', 'Owner'],
                         acquire=0)
    it.manage_permission(CMFCorePermissions.AccessFuturePortalContent,
                         roles=['TrackerAdmin','Supporter','Reporter', 'Manager', 'Owner'],
                         acquire=0)
    it.manage_permission(CMFCorePermissions.AddPortalContent,
                         roles=['TrackerAdmin','Supporter','Reporter', 'Manager', 'Owner'],
                         acquire=0)

    install_proxied_scripts(it)

    
    if REQUEST is not None:
        try:    url=self.DestinationURL()
        except: url=REQUEST['URL1']
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
    return id

def install_proxied_scripts(container):
    """ install Pythonscripts inside the ZODB. They need to be
        run with Manager proxy role.
    """

    script_dir = os.path.join(os.path.dirname(__file__), 'scripts')

    for sc in ('get_users', 'watchlist_action'):
        script = PythonScript(sc)
        script.write(open(os.path.join(script_dir, sc + '.py')).read())

        try:
            container._setObject(sc, script)
            getattr(container, sc)._proxy_roles = ('Manager',)
        except:
            traceback.print_exc()
