#############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Implement the CollectorNG Issue content type - a bundle containing the
collector transcript and various parts."""

import os, urllib, re, smtplib, traceback
from DateTime import DateTime
from Globals import InitializeClass
from AccessControl import ClassSecurityInfo, getSecurityManager
from Acquisition import aq_base
from Products.CMFDefault.DublinCore import DefaultDublinCoreImpl
from Products.CMFCore.WorkflowCore import WorkflowAction
from Products.CMFCore.utils import getToolByName

from Products.CMFDefault.SkinnedFolder import SkinnedFolder
from zLOG import LOG, ERROR

# Import permission names
from Products.CMFCore.CMFCorePermissions import View
from Products.CMFCore.CMFCorePermissions import ModifyPortalContent
from CollectorNGPermissions import *
from XML4CollectorNG import IssueXML

from Transcript import Transcript, TranscriptEntry
from transcript_formatter import format_transcript
from References import ReferencesManager, Reference
import util                             # CollectorNG utilities.
import sys, time, base64, random
from email.MIMEMultipart import MIMEMultipart
from email.MIMEText import MIMEText
from email.MIMEImage import MIMEImage
from email.Header import Header
import email.Utils 

factory_type_information = (
    {'id': 'CollectorNG Issue',
     'icon': 'collector_issue_icon.gif',
     'meta_type': 'CMF CollectorNG Issue',
     'description': ('A CollectorNG Issue represents a bug report or'
                     ' other support request.'),
     'product': 'CMFCollectorNG',
     'factory': None,                   # So not included in 'New' add form
     'allowed_content_types': ( 'File', 'Image'), 
     'immediate_view': 'collector_edit_form',
     'actions': ({'id': 'ticketbrowser',
                  'name': 'Browse',
                  'action': 'collector_contents',
                  'permissions': (ViewCollector,)},
                 {'id': 'view',
                  'name': 'View',
                  'action': 'collector_issue_contents',
                  'permissions': (ViewCollector,)},
                 {'id': 'followup',
                  'name': 'Followup',
                  'action': 'collector_issue_followup_form',
                  'permissions': (AddCollectorIssueFollowup,)},
                 {'id': 'ticketreferencesupload',
                  'name': 'References and Uploads',
                  'action': 'collector_issue_references_uploads',
                  'permissions': (AddCollectorIssueReferencesUploads,)},
                 {'id': 'tickethistory',
                  'name': 'History',
                  'action': 'collector_issue_transcript_form',
                  'permissions': (ViewCollector,)},
                 {'id': 'ticketedit',
                  'name': 'Edit',
                  'action': 'collector_issue_edit_form',
                  'permissions': (EditCollectorIssue,)},
                 {'id': 'addIssue',
                  'name': 'New Issue',
                  'action': 'collector_issue_add_issue',
                  'permissions': (AddCollectorIssue,)},
                 ),
     },
    )

class CollectorNGIssue(SkinnedFolder, DefaultDublinCoreImpl, IssueXML):
    """An individual support request in the CMF CollectorNG."""

    meta_type = 'CMF CollectorNG Issue'
    _collector_path = None
    
    __ac_roles__ = ('Manager', 'TrackerAdmin', 'Supporter','Reporter')
    ACTIONS_ORDER = ['Accept', 'Assign', 'Resolve', 'Reject', 'Defer', 'Resign'] 

    security = ClassSecurityInfo()

    def __init__(self,
                 id, container,
                 properties=None,
                 reference=None,
                 upload=None):

        # This must go away    
        self.portal_type = 'CollectorNG Issue'
        self.assignees  = None

        self.cookie = base64.encodestring(str(time.time()*random.random()))[:-1]

        self.transcript = Transcript()
        self.references = ReferencesManager()

        # check or add default values
        for prop in container.issuePropertyManager.getProperties():
            # avoid name clashes with status()
            if prop.getId() in ('status',): continue

            if not hasattr(properties, prop.getId()):
                setattr(properties, prop.getId(), prop.getDefaultValue())

        # check submitter_id
        mbtool = getToolByName(container, 'portal_membership')
        username = str(mbtool.getAuthenticatedMember())
        if not properties.submitter_id:
            properties.submitter_id = username

        # now set the attributes        
        for k in properties.keys():
            setattr(self, k, getattr(properties, k))

        SkinnedFolder.__init__(self, id, self.title)
        self._set_collector_path(container)

        # 'contained' is for stuff that needs collector acquisition wrapping.
        container._setObject(id, self)
        contained = container._getOb(id)
        contained._setPortalTypeName('CollectorNG Issue')

        # Make the Irish guys happy
        DefaultDublinCoreImpl.__init__(contained,
                                       title=self.title, 
                                       description=self.description)


    def _set_collector_path(self, collector):
        """Stash path to containing collector."""
        # For getting the internal catalog when being indexed - at which 
        # time we may not have an acquisition context...
        self._collector_path = "/".join(collector.getPhysicalPath())


    security.declareProtected(EditCollectorIssue, 'edit')
    def edit(self, properties):
        """Update the explicitly passed fields."""

        userid = str(getSecurityManager().getUser())
        t_entry = TranscriptEntry(userid)

        for item in properties.keys():
            old = getattr(self, item, '')
            new = getattr(properties, item)

            if old != new:
                t_entry.addChange(self.issuePropertyManager.getPropertyById(item).getDescription(), old, new)
                setattr(self, item, new) 

        self._validate_properties()
        
        self.transcript.add(t_entry)
        self.transcript._p_changed = 1

        self.reindexObject()
        self._send_update_notice('Edit')

        # update DublinCore    
        self.notifyModified()

        return t_entry


    def _validate_properties(self):

        missing = []    
        for prop in self.issuePropertyManager.getProperties():
            val = getattr(self, prop.getId())
            if prop.getMandatory() and (val is None or val==''):
                missing.append(prop.getDescription())

        if missing:
            raise ValueError, self.trans('The following fields must be filled: %s') % \
                ', '.join([ '"%s"' % x for x in missing])


    security.declareProtected(AddCollectorIssueFollowup, 'do_action')
    def do_action(self, action, comment, assignees=None, upload=None, reference=None, properties=None, notification=1):
        """Execute an action, adding comment to the transcript."""

        username = str(getSecurityManager().getUser())
        t_entry = TranscriptEntry(username)

        if properties:
            for item in properties.keys():
                old = getattr(self, item)
                new = getattr(properties, item)

                if old != new:
                    t_entry.addChange(self.issuePropertyManager.getPropertyById(item).getDescription(), old, new)
                    setattr(self, item, new) 

        if upload:
            file = upload.file
            fileid = upload.fileid
            filetype = upload.get('filetype', 'file')
            filecomment = upload.comment
        else:
            file = fileid = filetype = filecomment = None

        orig_supporters = self.assigned_to()

        # add transcript entry for changes

        if assignees:
            assignees = filter(None, assignees)

        if assignees is not None:
            t_entry.addChange(
                self.issuePropertyManager.getPropertyById('assignees').getDescription(), 
                orig_supporters, 
                assignees)                

        if comment: t_entry.addComment(comment)

        # references

        if reference:
            if reference.tracker and reference.ticketnumber:

                if not reference.comment:
                    raise ValueError,'References must have a comment'

                # check the reference
                tracker = getattr(self, reference.tracker, None)
                if not tracker:
                    raise ValueError, 'Tracker does not exist: %s' % reference.tracker

                # This code raises for some reasons an interal 
                # SystemError(error return without exception set)

                if getattr(tracker.aq_base, str(reference.ticketnumber), None) is None:
                    raise ValueError, 'Ticket number does not exist: %s' % reference.ticketnumber

                ref = Reference(reference.tracker, reference.ticketnumber, reference.comment)
                t_entry.addReference(reference.tracker, reference.ticketnumber,  reference.comment)
                self.references.add(ref)
                self.references._p_changed = 1

        # Strip off '_confidential' from status, if any.
        old_status = self.status()

        # if someone is changing the supporters list without setting the
        # the workflow step to "assign", we set the action to "assign" fo
        # usability reasons.

        if action.lower()=='comment' and not util.lists_eq(orig_supporters, assignees) and assignees is not None:
            action = 'assign'

        if action.lower() != 'comment':
            # Confirm against portal actions tool:
            if action != 'request' and action not in self.validActions():
                raise 'Unauthorized', "Invalid action '%s'" % action

            self.portal_workflow.doActionFor(self,
                                             action,
                                             comment=comment,
                                             username=username,
                                             assignees=assignees)

        new_status = self.status()

        if old_status != new_status:
            t_entry.addChange(self.issuePropertyManager.getPropertyById('status').getDescription(), 
                              old_status, new_status)

        # File upload
        fileid = self._process_file(file, fileid, filetype, filecomment)
        if fileid:
            t_entry.addUpload(fileid, filecomment)

        self.reindexObject()
                                       
        if notification:
            self.transcript.add(t_entry)
            self.transcript._p_changed = 1
            got = self._send_update_notice(action, new_status, file_id=fileid)
        else: got = ""

        self.notifyModified()
    
        return got


    def validActions(self):
        """Return actions valid according to workflow/application logic."""

        pa = getToolByName(self, 'portal_actions', None)
        allactions = pa.listFilteredActionsFor(self)
        return [entry['name']
                for entry in allactions.get('issue_workflow', [])]


    def _send_update_notice(self, action, new_status=None, file_id=None):
        """Send email notification about issue event to relevant parties."""

        if self.dispatching == 'none': return

        action = action.split('_')[0].capitalize()
        new_status = self.status()

        membership = getToolByName(self, 'portal_membership', None)

        userids = [ self.submitter_id ]
        userids.extend(self.managers)    # all managers

        if self.dispatching == 'assigned':
            userids.extend(self.assigned_to())   # all assigned guys
        else:
            # all supporters (only for non-confidential issues)
            if not self.confidential():
                userids.extend(self.supporters)

        # filter out recipients without email
        userids = util.remove_dupes(userids)

        rcp_with_attach = []
        rcp_without_attach = []

        for userid in userids:
            member = membership.getMemberById(userid)
            wants_attach = util.safeGetProperty(member, "submitter_send_attachments", "no")
            user_email = util.safeGetProperty(member, "email", "")
            
            if wants_attach == 'yes':
                rcp_with_attach.append(user_email)
            else:
                rcp_without_attach.append(user_email)

        # Watchers

        self._check_watchers()
        rcp_without_attach.extend(self.getWatchers())

        # special notications based on the state
        email = self.state_email.get(new_status, None)
        if email:
            rcp_without_attach.append(email)

        # add submitter email
        rcp_without_attach.append(self.submitter_email)

        self._send_out_notification(rcp_with_attach, action, file_id, 1)
        self._send_out_notification(rcp_without_attach, action, file_id, 0)


    def _send_out_notification(self, recipients, action, file_id, send_attachments=0):

        if len(recipients) == 0: return

        mfrom = self.email  # from CollectorNG
        subject = '[%s] #%s (%s): %s' % \
                  (self.abbrev, self.getId(), action,  self.title)

        subject = str(Header(subject, 'iso-8859-1'))

        body = ''
        email_submissions = getattr(self, 'email_submissions')
        if email_submissions:
            s = "\n\n*** INSERT YOUR REPLY BEFORE THIS LINE *** (%s) ***\n" % self.cookie
            s+= "*** DO NOT REMOVE THIS SECTION WHEN REPLYING TO THIS EMAIL ***\n\n"
            body += s  

        body +=  format_transcript(self) 

        outer = MIMEMultipart()
        outer['From'] = mfrom
        outer['To'] = ','.join(recipients)
        outer['Subject'] = subject
        outer['Content-Type'] = 'text/plain; charset=iso-8859-15'
        outer['Message-ID'] = email.Utils.make_msgid()
        outer['Reply-To'] = self.aq_parent.email 
        
        body_msg = MIMEText(body, _charset='iso-8859-15')
        outer.attach(body_msg)

        if send_attachments:
            if file_id:
                obj = getattr(self, file_id)
                if obj.meta_type == 'Portal Image':
                    outer.attach( MIMEImage(str(obj.data) ))

        try:        
            self.MailHost._send(mfrom, recipients, outer.as_string())
        except:
            LOG("CMFCollectorNG", ERROR, "Mailhost call failed", 
                error=sys.exc_info())


    def _process_file(self, file, fileid, filetype, comment):
        """Upload file to issue if it is substantial (has a name).

        Return a message describing the file, for transcript inclusion."""
        if file and file.filename:
            if not fileid:
                # This code sucks
                fileid = file.filename.split('/')[-1].split('\\')[-1]
            upload = self._add_artifact(fileid, filetype, comment, file)
            return fileid
        else:
            return None

    def _add_artifact(self, id, type, description, file):
        """Add new artifact, and return object."""
        self.invokeFactory(type, id)
        it = self._getOb(id)
        # Acquire view and access permissions from container
        it.manage_permission('View', acquire=1)
        it.manage_permission('Access contents information', acquire=1)
        it.description = description
        it.manage_upload(file)
        return it


    security.declareProtected(View, 'assigned_to')
    def assigned_to(self):
        """Return the current supporters list, according to workflow."""
        wftool = getToolByName(self, 'portal_workflow')
        return wftool.getInfoFor(self, 'assigned_to', ()) or ()

    security.declareProtected(View, 'status')
    def status(self):
        """Return the current status according to workflow."""
        wftool = getToolByName(self, 'portal_workflow')
        return wftool.getInfoFor(self, 'state', 'Pending')

    security.declareProtected(View, 'review_state')
    review_state = status

    security.declareProtected(View, 'confidential')
    def confidential(self):
        """True if workflow has the issue marked confidential.

        (Security_related issues start confidential, and are made
        unconfidential on any completion.)"""
        wftool = getToolByName(self, 'portal_workflow')
        return wftool.getInfoFor(self, 'confidential', 0)


    #################################################
    # Dublin Core and search provisions

    # The transcript indexes itself, we just need to index the salient
    # attribute-style issue data/metadata...

    security.declareProtected(ModifyPortalContent, 'indexObject')
    def indexObject(self):
        for i in (self._get_internal_catalog(),
                  getToolByName(self, 'portal_catalog', None)):
            if i is not None:
                i.indexObject(self)

    security.declareProtected(ModifyPortalContent, 'unindexObject')
    def unindexObject(self):
        for i in (self._get_internal_catalog(),
                  getToolByName(self, 'portal_catalog', None)):
            if i is not None:
                i.unindexObject(self)

    security.declareProtected(ModifyPortalContent, 'reindexObject')
    def reindexObject(self, idxs=[], internal_only=0):

        catalogs = [self._get_internal_catalog()]
        if not internal_only:
            catalogs.append(getToolByName(self, 'portal_catalog', None))

        for i in catalogs:
            if i is not None:
                i.reindexObject(self)

    def _get_internal_catalog(self):
        if self._collector_path is None:
            # Last ditch effort - this will only work when we're being called
            # from the collector's .reindex_issues() method.
            self._set_collector_path(self.aq_parent)
        container = self.restrictedTraverse(self._collector_path)
        return container.get_internal_catalog()

    def manage_afterAdd(self, item, container):
        """Add self to the workflow and catalog."""
        # Are we being added (or moved)?
        if aq_base(container) is not aq_base(self):
            self._set_collector_path(self.aq_parent)
            wf = getToolByName(self, 'portal_workflow', None)
            if wf is not None:
                wf.notifyCreated(self)
            self.indexObject()

    def manage_beforeDelete(self, item, container):
        """Remove self from the catalog."""
        # Are we going away?
        if aq_base(container) is not aq_base(self):
            self.unindexObject()
            # Now let our "aspects" know we are going away.
            for it, subitem in self.objectItems():
                si_m_bD = getattr(subitem, 'manage_beforeDelete', None)
                if si_m_bD is not None:
                    si_m_bD(item, container)

    security.declareProtected(View, 'is_assigned')
    def is_assigned(self):
        """True iff the current user is among .assigned_to()."""
        username = str(getSecurityManager().getUser())
        return username in (self.assigned_to() or ())


    security.declareProtected(ModifyPortalContent, 'setProperty')
    def setProperty(self, k,v):
	""" set a property """
	setattr(self, k, v)
	

    def SearchableText(self):
        """Consolidate all text and structured fields for catalog search."""
        # Make this a composite of the text and structured fields.

        l = []
        for attr in ('title' ,'description' ,'solution' ,'topic' ,'subtopic',
                     'classification' ,'importance' ,'status' ,'version_info',
                     'submitter_name' ,'submitter_city' ,'submitter_fax',
                     'submitter_phone' ,'submitter_address' ,'submitter_email'):

            v = getattr(self, attr, None)
            if v:
                if callable(v): v = v()
                l.append( str(v) )

        l.append( format_transcript(self) )
        l.append( str((self.security_related and 'security_related') or '') )

        return ' '.join(l)

    def Subject(self):
        """The structured attributes."""
        return (self.topic,
                self.classification,
                str(self.importance),
                )

    def __len__(self):
        """ number of transcript entries """
        return len(self.transcript)

    def __repr__(self):
        return ("<%s %s \"%s\" at 0x%s>"
                % (self.__class__.__name__,
                   self.id, self.title,
                   hex(id(self))[2:]))

    def getIssueTitle(self):
        """ return title of issue """
        return self.title

    def getIssueDescription(self):
        """ return description of issue """
        return self.description    

    def getTranscript(self):
        """ return the transcript """
        return self.transcript

    security.declareProtected(ModifyPortalContent, 'dump')
    def dump(self):
        """ return all attributes as text """

        lst = []

        for k in self.issuePropertyManager.getPropertyIds():
            v = getattr(self, k)
            if callable(v): v = v()
            lst.append('%-20s: %s (%s)' % (k, str(v), type(v)) )

        return '\n'.join(lst)


    def isVisibleFor(self, user=None):
        """ determine if this issue is visible for a user:
            issue is not confidential -> visible
            issue is confidential and member has either
            the TrackerAdmin, Manager or Owner role or
            is an assigned supporter -> visible
            else -> not visible
        """                    
        if user is None:
            user = getSecurityManager().getUser()

        # get user roles
        roles = user.getRolesInContext(self)

        # Issue is not confidential -> it is visible
        if not self.confidential(): return 1

        # Check roles that can always view confidential tickets
        for role in ('Manager', 'TrackerAdmin', 'Owner'):
            if role in roles: 
                return 1

        # assigned reporters may also see a confidential issue
        if 'Supporter' in roles and user.getUserName() in self.assigned_to():
            return 1

        return 0

    ###########################################################
    # Watchlist
    ###########################################################

    def _check_watchers(self):
        """ needed for backward compatibility """
        try: self._watchers
        except: self._watchers = []

    def getWatchers(self):
        self._check_watchers()
        return self._watchers

    def isWatcher(self, email):
        self._check_watchers()
        return email.lower() in self._watchers

    def addWatcher(self, email):
        self._check_watchers()
        
        if self.isWatcher(email):
            raise RuntimeError, 'Email address %s already registered' % email
        self._watchers.append(email.lower())


    def removeWatcher(self, email):
        self._check_watchers()
        if not self.isWatcher(email):
            raise RuntimeError, 'Email address %s is not registered' % email
        self._watchers.remove(email.lower())
    


InitializeClass(CollectorNGIssue)
    

def addCollectorNGIssue(self, id, properties, upload=None, reference=None, REQUEST=None):
    """Create a new issue in the collector.

    We return a string indicating any errors, or None if there weren't any."""

    it = CollectorNGIssue(id=id, container=self, properties=properties, 
                          upload=upload, reference=reference)
    it = self._getOb(it.id)
    it._validate_properties()
    it = it.__of__(self)

    it._addRole('Supporter')
    it._addRole('TrackerAdmin')
    it._addRole('Reporter')

    it.manage_permission(CMFCorePermissions.AddPortalContent,
                         roles=['TrackerAdmin','Supporter','Reporter', 'Manager', 'Owner'],
                         acquire=1)
    
    got = it.do_action('request', properties.description, properties.assignees, 
                       upload=upload, reference=reference)

    return got
